@@ -153,6 +153,9 @@ static void input_conn_free(struct input_conn *iconn)
 
 static void input_device_free(struct input_device *idev)
 {
+	if (idev->dc_id)
+		device_remove_disconnect_watch(idev->device, idev->dc_id);
+
 	dbus_connection_unref(idev->conn);
 	btd_device_unref(idev->device);
 	g_free(idev->name);
@@ -424,6 +427,7 @@ static gboolean intr_watch_cb(GIOChannel *chan, GIOCondition cond, gpointer data
 				"Connected", DBUS_TYPE_BOOLEAN, &connected);
 
 	device_remove_disconnect_watch(idev->device, idev->dc_id);
+	idev->dc_id = 0;
 
 	iconn->intr_watch = 0;
 
@@ -777,13 +781,17 @@ static int disconnect(struct input_device *idev, uint32_t flags)
 	return connection_disconnect(iconn, flags);
 }
 
-static void disconnect_cb(struct btd_device *device, void *user_data)
+static void disconnect_cb(struct btd_device *device, gboolean removal,
+				void *user_data)
 {
 	struct input_device *idev = user_data;
+	int flags;
 
 	info("Input: disconnect %s", idev->path);
 
-	disconnect(idev, 0);
+	flags = removal ? (1 << HIDP_VIRTUAL_CABLE_UNPLUG) : 0;
+
+	disconnect(idev, flags);
 }
 
 static int input_device_connected(struct input_device *idev,
@@ -1003,8 +1011,6 @@ static void device_unregister(void *data)
 	debug("Unregistered interface %s on path %s", INPUT_DEVICE_INTERFACE,
 								idev->path);
 
-	/* Disconnect if applied */
-	disconnect(idev, (1 << HIDP_VIRTUAL_CABLE_UNPLUG));
 	devices = g_slist_remove(devices, idev);
 	input_device_free(idev);
 }
